"""Copyright (c) 2017 Cisco Systems, Inc.
Name:
    asa.py
Usage:
    Library for legacy ASA command line.
Author:
    aitang/raywa
"""

import re
import time
import logging

from unicon.eal.dialogs import Dialog
from unicon.utils import AttributeDict


try:
    from kick.graphite.graphite import publish_kick_metric
except ImportError:
    from kick.metrics.metrics import publish_kick_metric

from .constants import AsaConfigConstants, AsaSmStates
from .patterns import AsaPatterns
from .statemachine import AsaStatemachine
from .dialogs import AsaDialog
from ...general.actions.basic import BasicDevice, BasicLine

LOGGER = logging.getLogger(__name__)

ASA_INIT_CMDS = """
    terminal width 511
    terminal pager 0
"""
DEFAULT_TIMEOUT = 10


class Asa(BasicDevice):
    """Class Asa.
    """

    def __init__(self, hostname='ciscoasa', enable_password=''):
        """Initializer of Asa instance.

        :param hostname: ASA hostname
        :param enable_password: ASA enable password
        :return: None

        """

        publish_kick_metric('device.asa.init', 1)
        self.patterns = AsaPatterns(hostname, enable_password)
        self.sm = AsaStatemachine(self.patterns)

        self.line_class = AsaLine
        super().__init__()

    def connect_from_ssp(self, ssp_line, slot, timeout=None, ld_name=''):
        """Connect through SSP telnet console

        :param ssp_line: SSP connection instance
        :param slot: module # that ASA is installed on
        :param timeout: timeout to wait for asa connection to be ready in seconds
        :param ld_name: logical device name for multi instance only
        :return: line object of ASA device

        To connect to ASA console from SSP MIO (BS-FPR9300/QP-FPR4100),
        following sequence of commands are needed:
         - connect module 1(2,3) console
           * we now enter Firepower mode, with prompts such as
             "Firepower-module1>","Firepower-module2>" or "Firepower-module3>"
         - connect asa
           * this brings us to asa mode with promps such as
             "ciscoasa>", "ciscoasa#" etc
             After this cli is entered, console messages from previous
             sessions will roll up. Sometimes these messages are not ended w/
             newline thus not able to bring up asa prompt.
             For this kind of scenarios, we need to send a newline to
             bring up asa prompt.
        """

        if not timeout:
            timeout = DEFAULT_TIMEOUT
        ssp_line.go_to('mio_state')

        # connect to module
        ssp_line.spawn_id.sendline('connect module {} console'.format(slot))
        time.sleep(0.5)
        ssp_line.spawn_id.sendline()
        ssp_line.spawn_id.sendline()

        # pylint: disable=anomalous-backslash-in-string
        asa_prompt = '[\r\n]({}|ciscoasa|asa)[/\w\(\)-]*[>#] '.format(self.patterns.hostname)
        dialog = Dialog([
            ['Firepower-module{}>'.format(slot),
             'sendline(connect asa {})'.format(ld_name), None, True, False],
            ['Connecting to ', 'sendline()', None, True, False],
            [asa_prompt, 'sendline()', None, False, False]
        ])
        dialog.process(ssp_line.spawn_id, timeout=timeout)

        asa_line = self.line_class(
            ssp_line.spawn_id, self.sm, 'telnet', carrier='ssp', timeout=timeout)
        return asa_line

    def connect_from_kp(self, kp_line, timeout=None):
        """Connect through SSP telnet console

        :param kp_line: KP connection instance
        :param timeout:
        :return: line object of ASA device

        """

        if not timeout:
            timeout = DEFAULT_TIMEOUT
        kp_line.go_to('fxos_state')

        # connect to asa
        kp_line.spawn_id.sendline('connect asa')

        asa_line = self.line_class(
            kp_line.spawn_id, self.sm, 'telnet', carrier='kp', timeout=timeout)
        return asa_line

    def connect_from_csp(self, csp_line, port, timeout=None):
        """

        :param csp_line:  CSP connection instance
        :param port: service port number
        :param timeout:
        :return:
        """
        csp_line.go_to('csp_enable_state')
        csp_line.spawn_id.sendline('telnet {port}'.format(port=port))
        csp_line.spawn_id.expect("Trying.*Connected to.*Escape character is '\^\]'\.", timeout)
        time.sleep(0.5)
        csp_line.spawn_id.sendline('')

        service_line = self.line_class(csp_line.spawn_id, self.sm, 'telnet', timeout=timeout)
        return service_line


class AsaLine(BasicLine):
    """ASA line class that provides connection interaction to the device
    """

    def __init__(self, spawn_id, sm, conn_type, carrier=None, timeout=None):
        """Initializer of AsaLine.

        :param spawn_id:
        :param sm: ASA statemachine
        :param conn_type: connection type (telnet or ssh)
        :return: None

        """

        super().__init__(spawn_id, sm, conn_type, timeout=timeout)
        self.carrier = carrier
        self.config(ASA_INIT_CMDS)

    def expect_prompt(self):
        """Expect current prompt
        """

        prompt = self.sm.get_state(self.sm.current_state).pattern
        self.spawn_id.sendline()
        return self.spawn_id.expect(prompt).last_match.string

    def change_hostname(self, new_hostname):
        """Change ASA hostname

        :param new_hostname: new hostname to be changed
        :return: None

        """

        patterns = self.sm.patterns
        patterns.change_hostname(new_hostname)
        self.sm.change_pattern(patterns)
        self.config('hostname %s' % new_hostname, exception_on_bad_config=True)

    def change_enable_password(self, new_password):
        """Change ASA enable password

        :param new_password: new enable password to be changed
        :return: None

        """

        patterns = self.sm.patterns
        patterns.change_enable_password(new_password)
        self.sm.change_pattern(patterns)
        self.config('enable password %s' % new_password, exception_on_bad_config=True)

    def delete(self, file_name, location=AsaConfigConstants.DISK0, flag=''):
        """Delete a file from asa

        Example:
            asa.delete('image.bin', 'disk0', '/noconfirm') will execute cmd:
            delete /noconfirm disk0: image.bin

        :param file_name: file to be deleted
        :param location: file location
        :param flag: flag to be used such as /noconfirm
        :return: None

        """

        if AsaConfigConstants.DELETE_NOCONFIRM in flag:
            self.enable_execute('delete %s %s:%s' % (flag, location, file_name))
        else:
            self.go_to(AsaSmStates.ENABLE_STATE.value)
            self.spawn_id.sendline('delete %s %s:%s' % (flag, location, file_name))
            self.spawn_id.expect('Delete filename')
            self.spawn_id.sendline()
            states = [state.name for state in self.sm.states]
            enable_prompt = self.sm.states[states.index(AsaSmStates.ENABLE_STATE.value)].pattern
            while True:
                output = self.spawn_id.read()
                if output:
                    if '[confirm]' in output:
                        self.spawn_id.sendline()
                    elif re.search(enable_prompt, output):
                        break
                time.sleep(0.5)

    def clear_config_all(self, timeout=60):
        """Clear all config in ASA

        :param timeout: timeout for prompt return
        :return: None

        """

        if 'telnet' in self.type:
            self.config('clear configure all', timeout)
            self.config(ASA_INIT_CMDS)
        elif 'ssh' in self.type:
            raise RuntimeError('Clear configure all under ssh will lose current connection to ASA')

    def reload(self, timeout=300, write_mem=True):
        """Reload ASA

        :param timeout: reload timeout for prompt return
        :return: None

        """

        # write memory before reload
        if write_mem:
            self.enable_execute('write memory', exception_on_bad_command=True)

        # reload ASA
        self.spawn_id.sendline('reload')
        # pylint: disable=anomalous-backslash-in-string
        self.spawn_id.expect('\[confirm\]')
        self.spawn_id.sendline()

        self._reload_expect_prompt(timeout=timeout)

        self.config(ASA_INIT_CMDS)

    def change_firewall_mode(self, fmode):
        """Change firewall mode between router and transparent

        :param fmode: firewall mode. tfw | rfw
        :return: None

        """

        # pylint: disable=anomalous-backslash-in-string
        show_mode = re.search('Firewall mode: (\w+)', self.enable_execute('show firewall'))
        if show_mode:
            current_mode = show_mode.group(1)
            if AsaConfigConstants.FIREWALL_MODE[fmode] != current_mode:
                cmd = 'firewall %s' % AsaConfigConstants.FIREWALL_MODE['tfw'].lower()
                if fmode == 'tfw':
                    self.config(cmd)
                elif fmode == 'rfw':
                    self.config('no ' + cmd)
                self.change_hostname(self.sm.patterns.hostname)
        else:
            raise RuntimeError('Unable to determine current firewall mode')

    def change_context_mode(self, cmode, timeout=300):
        """Change context mode between single and multi

        :param cmode: context mode. sfm | mfm
        :return: None

        """

        # pylint: disable=anomalous-backslash-in-string
        show_mode = re.search('Security context mode: (\w+)', self.enable_execute('show mode'))
        if show_mode:
            current_mode = show_mode.group(1)
        else:
            raise RuntimeError('Unable to determine current context mode')

        if current_mode == AsaConfigConstants.CONTEXT_MODE['mfm']:
            self.enable_execute('changeto system')

        if AsaConfigConstants.CONTEXT_MODE[cmode] != current_mode:
            self.go_to(AsaSmStates.CONFIG_STATE.value)
            self.spawn_id.sendline('mode %s' % AsaConfigConstants.CONTEXT_MODE[cmode])
            # pylint: disable=anomalous-backslash-in-string
            self.spawn_id.expect('change mode\? \[confirm\]')
            self.spawn_id.sendline()
            try:
                self.spawn_id.expect('system configuration\? \[confirm\]')
            except:
                pass
            else:
                self.spawn_id.sendline()

            self._reload_expect_prompt(timeout=timeout)

            self.config(ASA_INIT_CMDS)

    def download_image(
            self, protocol, server, image_path, dst,
            port=None, noconfirm=True, overwrite=True, usr=None, pwd=None, timeout=600):
        """Download a required image from the source to the destination

        :param protocol: protocol going to be used
        :param server: server address
        :param image_path: image path
        :param dst: destination path on ASA
        :param noconfirm: flag to skip download dialog
        :param usr: username
        :param pwd: password
        :return: None

        """
        cmd = 'copy '
        if noconfirm:
            cmd += '/noconfirm '

        cmd += '{protocol}://'.format(protocol=protocol)

        if usr:
            cmd += '{usr}:{pwd}@'.format(usr=usr, pwd=pwd)

        cmd += '{server}'.format(server=server)

        if port:
            cmd += ':{port}'.format(port=port)

        cmd += '/{image_path} {dst}'.format(image_path=image_path, dst=dst)

        if noconfirm:
            self.execute(cmd, timeout=timeout, exception_on_bad_command=True)
        else:
            overwrite = '' if overwrite is True else 'q'
            args = AttributeDict({'overwrite': overwrite})

            self.spawn_id.sendline(cmd)
            dialog = AsaDialog.download_dialog
            output = dialog.process(self.spawn_id, context=args, timeout=timeout)

            if 'Error' in output.last_match.string:
                raise RuntimeError('bad command: {cmd}'.format(cmd=cmd))

    def disconnect(self):
        """Disconnect the line.

        :return: None

        """
        if self.spawn_id is not None:
            if self.type == 'ssh':
                # send \n + ~.
                self.spawn_id.sendline('')
                self.spawn_id.send('~.')
                self.spawn_id.expect('Connection to .* closed.')
            elif self.type == 'telnet':
                if self.carrier in ('ssp', 'kp'):
                    # pylint: disable=anomalous-backslash-in-string
                    self.spawn_id.send('\001')
                    self.spawn_id.send('d')
                    if self.carrier == 'ssp':
                        self.spawn_id.expect('Firepower-module\d+>')
                        self.spawn_id.send('~')
                        self.spawn_id.expect('telnet>')
                        self.spawn_id.sendline('quit')
                        self.spawn_id.expect('Connection closed.')
                else:
                    # send ctrl + ], then q
                    self.spawn_id.send('\035')
                    self.spawn_id.expect('telnet> ')
                    self.spawn_id.sendline('q')
                    self.spawn_id.expect('Connection closed.')
            self.spawn_id.close()
        else:
            LOGGER.info('You have already closed the connection to this device previously.')
        self.spawn_id = None

    def _reload_expect_prompt(self, timeout=300):
        """Wait for ASA to be back online after reload

        :param timeout: timeout for reload
        :return: None

        """

        if 'telnet' in self.type:
            disable_prompt = self.sm.patterns.prompt.disable_prompt

            dialog = Dialog([
                ['Process shutdown finished', None, None, True, False],
                ['The system is restarting', None, None, True, False],
                ['Configuring network interfaces... done', None, None, True, False],
                ['Compiled on', None, None, True, False],
                ['The digital signature of the running image verified successfully', \
                 None, None, True, False],
                ['Compiled on', None, None, True, False],
                ['Cisco Adaptive Security Appliance Software Version', None, None, True, False],
                [disable_prompt, 'sendline()', None, False, False],
            ])
            dialog.process(self.spawn_id, timeout=timeout)

            self.go_to('any')
        elif 'ssh' in self.type:
            pass

    def sudo_execute(self, cmd, timeout=None, exception_on_bad_command=False):
        """ Not implemented for AsaLine class"""

        raise NotImplementedError



"""patterns.py.

Chassis statements

"""

from unicon.eal.dialogs import Statement


def login_handler(spawn, username, password):
    spawn.sendline(username)
    spawn.expect('Password: ')
    spawn.sendline(password)
    spawn.expect('Cisco Firepower')
    spawn.sendline()


class ChassisStatements:
    def __init__(self, patterns, chassis_data):
        username = \
            chassis_data['custom']['chassis_login']['username']
        password = \
            chassis_data['custom']['chassis_login']['password']
        self.login_password = \
            Statement(
                pattern=patterns.prompt.prelogin_prompt,
                action=login_handler,
                args={
                    'username': username,
                    'password': password
                },
                loop_continue=True, continue_timer=True)





def publish_kick_metric(
        metric_name,
        value,
        moment=None,
        user=None,
        host_name=None):
    """Publish a KICK!

    metric based on its name and value. If moment is not set, use the
    current time. If user is not set, determine it with _get_username().
    If host_name is not set, determine it with __get_host_hostname().
    The structure of the resulting metric:
    kick.<user>.<host_name>.<metric_name>

    """
    pass



import munch


class Ftd5500xPatterns:
    def __init__(self, hostname, login_password, sudo_password, enable_password):
        self._hostname = hostname
        self.username = 'admin'
        self.login_password = login_password
        self.sudo_password = sudo_password
        self.enable_password = enable_password
        self.default_password = 'Admin123'
        self.prompt = munch.Munch()
        self.set_prompt_patterns()

    @property
    def hostname(self):
        return self._hostname

    @hostname.setter
    def hostname(self, value):
        self._hostname = value
        self.set_prompt_patterns()

    def set_prompt_patterns(self):

        self.prompt.password_prompt = '\r\n[Pp]assword: $'
        self.prompt.prelogin_prompt = '\r?\n?({}|firepower|ciscoasa) login: $'.format(self.hostname)
        self.prompt.fireos_prompt = r'\r\n(\x1bE\x1b\[J)?[\x07]?> $'
        self.prompt.expert_prompt = r'\r\n(\x1b.{{1,10}})?(admin@)?({}|firepower|ciscoasa):.*?\$ '.format(self.hostname)
        self.prompt.system_prompt = '\r\n[\x07]?system> $'
        self.prompt.sudo_prompt = '\r\n[^\r\n]*(root@.*#) ( \r|\r )*$'
        self.prompt.disable_prompt = '[\r\n]({}|firepower)> $'.format(self.hostname)
        self.prompt.enable_prompt = '[\r\n]({}|firepower)# $'.format(self.hostname)
        self.prompt.ciscoasa_prompt = '\r\nciscoasa:~\$'
        self.prompt.rommon_prompt = 'rommon.*> $'
        self.prompt.firepower_boot = '-boot>'
        self.prompt.config_prompt = '[\r\n]({}|firepower)[\w]*\([\w\-]+\)# $'.format(self.hostname)





"""Library to load new image on an Elektra 5585 SFR blade.

We require a file server to be co-located with the device. This file
server has both the rommon_image and the pkg_image mounted.
Elektra will first break into rommon, tftp the
rommon_image and boot from it into Elektra boot CLI. Then it will do an HTTP
copy of pkg_image to boot up into a fully functional device. This second
step takes a long time (~30 min), as it will run a lot of scripts.

The file server is a PXE server in each site, defined in gcp_constants.py.

"""
import re
import time
import logging
logger = logging.getLogger(__name__)

from unicon.eal.expect import TimeoutError
from unicon.eal.dialogs import Dialog
from ...general.actions.basic import BasicDevice, BasicLine
from ...general.actions.power_bar import power_cycle_all_ports

try:
    import kick.graphite.graphite as graphite
except ImportError:
    import kick.metrics.metrics as graphite

MAX_RETRY_COUNT = 3


class Elektra5585(BasicDevice):
    def __init__(self, hostname='firepower', login_password='Admin123',
                 sudo_password='Admin123', enable_password='', *args, **kwargs):
        """Constructor for Elektra5585.

        Add *args and **kwargs so that the initiator can be invoked with
        additional (unused) arguments.

        :param hostname: host name as appears in prompt
                e.g. 'BATIT-5508-3-FUL'
        :param login_password: password for login
        :param sudo_password: root password
        :param enable_password: password for privileged exec mode
        :return: None

        """

        super().__init__()

        graphite.publish_kick_metric('device.elektra5585.init', 1)
        from .patterns import Elektra5585Patterns
        self.patterns = Elektra5585Patterns(hostname, login_password, sudo_password, enable_password)

        # create the state machine that contains the proper attributes.
        from .statemachine import Elektra5585Statemachine
        self.sm = Elektra5585Statemachine(self.patterns)

        # important: set self.line_class so that a proper line can be created
        # by ssh_console(), etc.
        self.line_class = Elektra5585Line


class Elektra5585Line(BasicLine):
    def __init__(self, spawn, sm, type, timeout=None):
        """ Constructor of Elektra5585

        :param spawn: spawn_id of the line
        :param sm: object of statemachine
        :param type: line type, e.g. 'ssh', 'telnet'
        :return: object of the line

        """

        try:
            super().__init__(spawn, sm, type, timeout)
        except:
            # Handle boot loop
            pass
            logger.info('=== Failed to go_to("any") '
                        'try to drop device to rommon.')
            count = 0
            while count < MAX_RETRY_COUNT:
                try:
                    self.rommon_go_to()
                    self.sm.update_cur_state('rommon_state')
                    logger.info('=== Dropped device to rommon.')
                    break
                except:
                    pass
                    count += 1
            assert count < MAX_RETRY_COUNT, '=== Could not drop device ' \
                                            'to rommon, assert ...'

    def wait_until_device_on(self, timeout=600):
        """ Wait a given period of time for device to be on

        :param timeout: wait time for the device to boot up
        :return: None

        """

        # The system will reboot, wait for the following prompts
        d1 = Dialog([
            ['SW-DRBG health test passed', None, None, False, False],
        ])
        d1.process(self.spawn_id, timeout=timeout)

        # sleep 60 seconds to avoid errors
        time.sleep(60)
        self.sm.go_to('any', self.spawn_id)

    def power_cycle(self, power_bar_server, power_bar_port,
                    wait_until_device_is_on=True, timeout=600,
                    power_bar_user='admn', power_bar_pwd='admn'):
        """reboots a device from a Power Data Unit equipment.

        :param power_bar_server: comma-separated string of IP addresses of the PDU's
        :param power_bar_port: comma-separated string of power port on the PDU's
        :param wait_until_device_is_on: True if waits for the device
                                to boot up, False otherwise
        :param timeout: wait time for the device to boot up
        :param power_bar_user: comma-separated string of usernames for power bar servers
        :param power_bar_pwd: comma-separated string of passwords for power bar servers
        :return: status of power cycle result, True or False

        """

        if power_bar_server == "" or power_bar_port == "":
            return None

        result = power_cycle_all_ports(power_bar_server, power_bar_port, power_bar_user, power_bar_pwd)

        logger.info('Wait for device to be up running ...')
        if wait_until_device_is_on:
            self.wait_until_device_on(timeout=timeout)
        return result

    # the following functions are for device baselining. This only works on
    # a console line (ssh or telnet).
    def expect_and_sendline(self, this_spawn, es_list, timeout=10):
        """takes a list of expect/send actions and perform them one by one.

        es_list looks like:
        [['exp_pattern1', 'send_string1', 30],
         ['exp_pattern2', 'send_string2'],
         ...
        ]

        The third element is for timeout value, and can be ommitted when the
        overall timeout value applies.

        :param this_spawn: the spawn of the line
        :param es_list: pairs of expected output and command to be sent
        :param timeout: default timeout for the command if timeout is not set
        :return: None

        """

        for es in es_list:
            if len(es) == 2:
                exp_pattern = es[0]
                send_string = es[1]
                to = timeout
            elif len(es) == 3:
                exp_pattern = es[0]
                send_string = es[1]
                to = int(es[2])
            else:
                raise RuntimeError("Unknown expect_and sendline input")

            this_spawn.sendline(send_string)
            this_spawn.expect(exp_pattern, timeout=to)
            time.sleep(3)

    def rommon_go_to(self):
        """Go to rommon mode.

        :return: None

        """

        # break into rommon
        d1 = Dialog([
            ['Use SPACE to begin boot immediately.',
             'sendline({})'.format(chr(27)), None, False, False],
        ])
        d1.process(self.spawn_id, timeout=900)
        time.sleep(10)
        logger.info('In rommon mode')

    def rommon_configure(self):
        """In ROMMON mode, set network, tftp info and rommon image For Kenton.

        :return: None

        """

        logger.info('add rommon config')
        es_list = [
            ['rommon', 'ADDRESS={}'.format(self.uut_ip)],
            ['rommon', 'SERVER={}'.format(self.rommon_tftp_server)],
            ['rommon', 'IMAGE={}'.format(self.rommon_image)],
            ['rommon', 'GATEWAY={}'.format(self.uut_gateway)],
            ['rommon', 'PORT=Management0/0'],
        ]

        self.expect_and_sendline(self.spawn_id, es_list)

    def rommon_boot(self, timeout=600):
        """Perform TFTP Boot from ROMMON.

        :return: None

        """

        time.sleep(10)
        logger.info('=== Performing TFTP Boot from ROMMON')
        logger.info('=== Wait for installation to complete, timeout = {}'
                    ' seconds ...'.format(str(timeout)))
        self.spawn_id.sendline('tftpdnld')
        d1 = Dialog([
            ['rommon.*> ', 'sendline(tftpdnld)', None, True, False],
            ['Overwrite \(y/n\)', 'sendline(y)', None, True, False],
            ['login:', 'sendline()', None, False, False],
        ])
        try:
            d1.process(self.spawn_id, timeout=timeout)
            self.spawn_id.sendline()
        except:
            logger.info("=== Rommon file download failed, raise runtime error. ")
            raise RuntimeError(
                ">>>>>> Download failed. Please check details - "
                "tftp_server: {}, image file: {}".format(self.rommon_tftp_server,
                                                         self.rommon_image))

        logger.info("=== Rommon file was installed successfully.")

    def firepower_boot_configure(self):
        """In boot CLI mode, set device network info, DNS server and domain.

        :return: None

        """

        d1 = Dialog([
            ['login:', 'sendline(admin)', None, True, False],
            ['Password: ', 'sendline({})'.format(self.sm.patterns.default_password),
             None, True, False],
            ['-boot>', 'sendline(setup)', None, False, False],
        ])
        d1.process(self.spawn_id, timeout=30)
        set_ntp_server = 'n' if self.ntp_server == None else 'y'
        d2 = Dialog([
            ['Enter a hostname ', 'sendline(firepower)', None, True, False],
            ['Do you want to configure IPv4 address on management interface',
             'sendline(y)', None, True, False],
            ['Do you want to enable DHCP for IPv4 address assignment on '
             'management interface', 'sendline(n)', None, True, False],
            ['Enter an IPv4 address', 'sendline({})'.format(self.uut_ip),
             None, True, False],
            ['Enter the netmask', 'sendline({})'.format(self.uut_netmask),
             None, True, False],
            ['Enter the gateway', 'sendline({})'.format(self.uut_gateway),
             None, True, False],
            ['Do you want to configure static IPv6 address on management '
             'interface', 'sendline(n)', None, True, False],
            ['Enter the primary DNS server IP address',
             'sendline({})'.format(self.dns_server.split(',')[0]), None, True, False],
            ['Do you want to configure Secondary DNS Server', 'sendline(n)',
             None, True, False],
            ['Do you want to configure Local Domain Name', 'sendline(n)',
             None, True, False],
            ['Do you want to configure Search domains', 'sendline(y)', None,
             True, False],
            ['Enter the comma separated list for search domains',
             'sendline({})'.format(self.search_domains), None, True, False],
            ['Do you want to enable the NTP service', 'sendline({})'.format(set_ntp_server),
             None, True, False],
            ['Enter the NTP servers separated by commas', 'sendline({})'.format(self.ntp_server),
             None, True, False],
            ['Do you want to enable the NTP symmetric key authentication', 'sendline(n)',
             None, True, False],
            ['Apply the changes', 'sendline(y)', None, True, False],
            ['Press ENTER to continue...', 'sendline()', None, True, False],
            ['-boot>', 'sendline()', None, False, False],
        ])
        d2.process(self.spawn_id, timeout=120)

    def firepower_install(self):
        """Perform ping test and verify the network connectivity to TFTP server.
        Install Elektra pkg image Enter device network info, hostname, and firewall
        mode.

        :return: None

        """

        for i in range(20):
            self.spawn_id.sendline('ping -c 1 {}'.format(self.rommon_tftp_server))
            try:
                self.spawn_id.expect('64 bytes from', timeout=5)
            except TimeoutError:
                time.sleep(60)
                continue
            else:
                break
        else:
            raise RuntimeError(">>>>>> Ping not working")

        d0 = Dialog([
            ['-boot>', 'sendline(system install {})'.format(self.pkg_image),
                None, False, False]
        ])

        d1 = Dialog([
            ['Do you want to continue?', 'sendline(y)', None, True, False],
            ['Upgrade aborted', 'sendline()', None, False, False],
            ['Installation aborted', 'sendline()', None, False, False]
        ])
        count = 0
        while count < self.retry_count:
            try:
                d0.process(self.spawn_id, timeout=20)
                d1.process(self.spawn_id, timeout=60)
                count += 1
                time.sleep(5)
            except:
                break
        assert count < self.retry_count, 'elektra installation failed' \
            ', please check elektra package url: "{}"'.format(self.pkg_image)

        d2 = Dialog([
            ['Do you want to continue with upgrade?', 'sendline(y)', None,
             True, True],
            ["Press 'Enter' to reboot the system.", 'sendline()', None, True,
             True],
            ['Use SPACE to begin boot immediately.', 'send(" ")', None, True,
             True],
            ['firepower login: ', 'sendline()', None, False, False],
        ])
        d2.process(self.spawn_id, timeout=3900)

        # Allow install processes to finish
        time.sleep(180)

        d3 = Dialog([
            ['firepower login: ', 'sendline(admin)', None, True, False],
            ['Password: ', 'sendline({})'.format(self.sm.patterns.default_password),
             None, True, False],
            ['Press <ENTER> to display the EULA: ', 'sendline()', None, True, False],
            ['--More--', 'send(q)', None, True, False],
            ["Please enter 'YES' or press <ENTER> to AGREE to the EULA: ",
             'sendline(YES)', None, False, False],
        ])
        d3.process(self.spawn_id, timeout=600)

        d4 = Dialog([
            ['firepower login: ', 'sendline(admin)', None, True, False],
            ['Password: ', 'sendline({})'.format(self.sm.patterns.login_password),
             None, True, False],
            ['Enter new password:', 'sendline({})'.format(self.sm.patterns.login_password),
             None, True, True],
            ['Confirm new password:', 'sendline({})'.format(self.sm.patterns.login_password),
             None, True, False],
            ['Do you want to configure IPv4', 'sendline(y)', None, True, False],
        ])

        if self.uut_ip6 is None:
            d4.append(['Do you want to configure IPv6', 'sendline(n)', None, True, False])
        else:
            d4.append(['Do you want to configure IPv6', 'sendline(y)', None, True, False])
        d4.append(['Configure IPv4 via DHCP or manually', 'sendline(manual)', None,
                   True, False])
        d4.append(['Enter an IPv4 address for the management interface',
                   'sendline({})'.format(self.uut_ip), None, True, False])
        d4.append(['Enter an IPv4 netmask for the management interface',
                   'sendline({})'.format(self.uut_netmask), None, True, False])
        d4.append(['Enter the IPv4 default gateway for the management interface',
                   'sendline({})'.format(self.uut_gateway), None, True, False])
        if self.uut_ip6 is not None:
            d4.append(['Configure IPv6 via DHCP, router, or manually',
                       'sendline(manual)', None, True, False])
            d4.append(['Enter the IPv6 address for the management interface',
                       'sendline({})'.format(self.uut_ip6), None, True, False])
            d4.append(['Enter the IPv6 address prefix for the management interface',
                       'sendline({})'.format(self.uut_prefix), None, True, False])
            d4.append(['Enter the IPv6 gateway for the management interface',
                       'sendline({})'.format(self.uut_gateway6), None, True, False])
        d4.append(['Enter a fully qualified hostname for this system ',
                   'sendline({})'.format(self.hostname), None, True, False])
        d4.append(['Enter a comma-separated list of DNS servers or',
                   'sendline({})'.format(self.dns_server), None, True, False])
        d4.append(['Enter a comma-separated list of search domains or',
                   'sendline({})'.format(self.search_domains), None, True, False])
        d4.append(['> ', 'sendline()', None, False, False])
        d4.process(self.spawn_id, timeout=900)

        logger.info('fully installed.')

    def configure_manager(self):
        """Configure manager to be used for registration

        :return: None

        """

        self.go_to('fireos_state')
        if self.manager_nat_id is None:
            response = self.execute('configure manager add {} {}'.format(self.manager,
                                                                         self.manager_key),
                                    120)
        else:
            response = self.execute('configure manager add {} {} {}'.format(self.manager,
                                                                            self.manager_key,
                                                                            self.manager_nat_id),
                                    120)
        success = re.search('Manager successfully configured', response)
        if success is None:
            logger.error('Exception: failed to configure the manager')
            raise RuntimeError('>>>>>> configure manager failed:\n{}\n'.format(response))

    def validate_version(self):
        """Checks if the installed version matches the version in package url

        :return: None

        """

        self.go_to('fireos_state')

        # check version
        response = self.execute('show version', 30)
        if not response:
            response = self.execute('show version', 30)
        build = re.findall('Build\s(\d+)', response)[0]
        version = re.findall(r'(Version\s){1}([0-9.]+\d)', str(response))[0][1]
        package_file = self.pkg_image.split('/')[-1]
        if (build in package_file) and (version in package_file):
            logger.info('>>>>>> show version result:\n{}\nmatches '
                        'elektra package image: {}'.format(response,
                                                           package_file))
            logger.info('Installed elektra version validated')
        else:
            logger.error('Exception: not the same version and build')
            raise RuntimeError('>>>>>> show version result:\n{}\ndoes not match '
                               'elektra package image: {}'.format(response,
                                                                  package_file))

    def baseline_elektra_5585(self, rommon_tftp_server, pkg_image,
                              uut_ip, uut_netmask, uut_gateway, rommon_image, dns_server,
                              hostname='firepower', search_domains='cisco.com',
                              retry_count=MAX_RETRY_COUNT,
                              power_cycle_flag=False, pdu_ip='', pdu_port='',
                              pdu_user='admn', pdu_pwd='admn', ntp_server=None,
                              uut_ip6=None, uut_prefix=None, uut_gateway6=None,
                              manager=None, manager_key=None, manager_nat_id=None, **kwargs):
        """Install rommon image and Elektra pkg image on SFR blade.

        :param rommon_tftp_server: TFTP Server IP Address
        :param pkg_image: Elektra image to be transferred via HTTP,
            e.g. 'http://192.168.0.50/Release/6.2.3-10697/installers/asasfr-sys-6.2.3-10697.pkg'
        :param uut_ip: Device IP Address to access TFTP Server
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param rommon_image: boot image under /tftpboot to be transferred via TFTP,
            e.g. 'netboot/elektra-boot-images/6.2.3/6.2.3-2/asasfr-boot-6.2.3-2.img'
        :param dns_server: DNS server
        :param hostname: hostname to be set
        :param search_domains: search domains delimited by comma,
            defaulted to 'cisco.com'
        :param retry_count: download retry count, defaulted to MAX_RETRY_COUNT
        :param power_cycle_flag: True power cylce before baseline, False otherwise
        :param pdu_ip: PDU IP
        :param pdu_port: PDU Port
        :param pdu_user: PDU admn
        :param pdu_pwd: PDU pwd
        :param ntp_server: NTP server delimited by comma, defaulted to None,
            otherwise the value of ntp_server is - e.g. "ntp.esl.cisco.com"
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param manager: FMC to be configured for registration
        :param manager_key: Registration key
        :param manager_nat_id: Registration NAT Id
        :param timeout: in seconds; time to wait for fetching the boot image from TFTP server;
                        defaulted to 600s
        :return: None

        """

        logger.info('Starting baseline')
        graphite.publish_kick_metric('device.elektra5585.baseline', 1)
        self.rommon_tftp_server = rommon_tftp_server
        self.pkg_image = pkg_image
        self.uut_ip = uut_ip
        self.uut_netmask = uut_netmask
        self.uut_gateway = uut_gateway
        self.uut_ip6 = uut_ip6
        self.uut_prefix = uut_prefix
        self.uut_gateway6 = uut_gateway6
        self.rommon_image = rommon_image
        self.dns_server = dns_server
        self.hostname = hostname
        self.search_domains = search_domains
        self.retry_count = retry_count
        self.ntp_server = ntp_server
        self.manager = manager
        self.manager_key = manager_key
        self.manager_nat_id = manager_nat_id
        tftptimeout = kwargs.get('timeout', 600)

        if not (self.sm.current_state in ['rommon_state', 'boot_state', 'pkglogin_state']):
            if not power_cycle_flag:
                logger.info('Reboot the device ...')
                self.go_to('sudo_state')
                self.spawn_id.sendline('reboot')
            else:
                logger.info('Power cycle the device ...')
                self.power_cycle(pdu_ip, pdu_port, wait_until_device_is_on=False, power_bar_user=pdu_user,
                                 power_bar_pwd=pdu_pwd)
            try:
                self.spawn_id.expect('Use (.*?BREAK.*?|.*?ESC.*?) to interrupt boot', timeout=120)
            except TimeoutError:
                RuntimeError(">>>>>> Failed to stop rebooting")
            logger.info('Drop the device to rommon.')
            self.rommon_go_to()

        if self.sm.current_state is 'pkglogin_state':
            self.go_to('boot_state')

        if self.sm.current_state is 'boot_state':
            logger.info('Device is in boot_state, drop the device to rommon.')
            self.spawn_id.sendline('system reload')
            d1 = Dialog([
                ['Are you sure you want to reload the system',
                 'sendline(y)', None, False, False],
            ])
            d1.process(self.spawn_id, timeout=30)
            self.rommon_go_to()

        logger.info('Rommon configure.')
        self.rommon_configure()

        logger.info('tftpdnld - tftp server: {}, '
                    'rommon image: {} ...'.format(rommon_tftp_server, rommon_image))
        self.rommon_boot(tftptimeout)
        self.go_to('any')
        logger.info('firepower boot configure ...')
        self.firepower_boot_configure()
        logger.info('Install image: {} ...'.format(pkg_image))
        self.firepower_install()
        self.go_to('any')
        if self.manager is not None:
            logger.info('Configure manager')
            self.configure_manager()
        logger.info('Validate version installed')
        self.validate_version()
        logger.info('Installation completed successfully.')



from unicon.eal.dialogs import Dialog
from unicon.statemachine import State, Path, StateMachine

from .dialogs import FmcDialogs
from .statements import FmcStatements


class FmcStatemachine(StateMachine):
    def __init__(self, patterns):
        """ Constructor of state machine for FMC
        """
        self.patterns = patterns
        self.dialogs = FmcDialogs(self.patterns)
        self.statements = FmcStatements(self.patterns)
        super().__init__()

    def create(self):
        # Create States and their state patterns
        prelogin_state = State('prelogin_state', self.patterns.prompt.prelogin_prompt)
        admin_state = State('admin_state', self.patterns.prompt.admin_prompt)
        sudo_state = State('sudo_state', self.patterns.prompt.sudo_prompt)
        fireos_state = State('fireos_state', self.patterns.prompt.fireos_prompt)
        liloboot_state = State('liloboot_state', self.patterns.prompt.lilo_boot_prompt)
        liloos_state = State('lilos_state', self.patterns.prompt.lilo_os_prompt)
        lilo_boot_menu_state = State('lilobootmenu_state', self.patterns.prompt.lilo_boot_menu_prompt)

        # Add our states to the state machine
        self.add_state(admin_state)
        self.add_state(sudo_state)
        self.add_state(prelogin_state)
        self.add_state(fireos_state)
        self.add_state(liloboot_state)
        self.add_state(liloos_state)
        self.add_state(lilo_boot_menu_state)

        # Create paths for switching between states
        admin_to_sudo = Path(admin_state, sudo_state, 'sudo su -',
                             self.dialogs.d_admin_to_sudo)

        sudo_to_admin = Path(sudo_state, admin_state, "exit", None)

        prelogin_to_admin = Path(prelogin_state, admin_state, 'admin',
                                 Dialog([self.statements.login_password_statement]))

        admin_to_prelogin = Path(admin_state, prelogin_state, 'exit', None)
        ##New added state
        prelogin_to_fireos_path = Path(prelogin_state, fireos_state, 'admin',
                                       Dialog([self.statements.login_password_statement]))
        ###End

        # transitions from and to 'fireos_state' state
        fireos_to_admin_path = Path(fireos_state, admin_state, 'expert', None)
        admin_to_fireos_path = Path(admin_state, fireos_state, 'exit', None)
        fireos_to_prelogin = Path(fireos_state, prelogin_state, 'exit', None)


        # Add paths to the State Machine
        self.add_path(admin_to_sudo)
        self.add_path(sudo_to_admin)
        self.add_path(prelogin_to_admin)
        self.add_path(admin_to_prelogin)
        self.add_path(fireos_to_admin_path)
        self.add_path(admin_to_fireos_path)
        self.add_path(fireos_to_prelogin)
        self.add_path(prelogin_to_fireos_path)

        # after inactivity timer, it will go back to prelogin:
        self.add_default_statements([self.statements.login_password, self.statements.login_incorrect])



"""patterns.py.

SSP prompt patterns

"""
import logging

import munch

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class SspPatterns:
    """SSP class that restores all prompt patterns."""
    def __init__(self, hostname, login_username, login_password, sudo_password, slot_id, app_hostname,
                 deploy_type='native', app_identifier=''):
        """Initializer of SspPatterns."""

        self.dev_hostname = hostname
        self.deploy_type = deploy_type
        self.app_identifier = app_identifier
        self.login_username = login_username
        self.login_password = login_password
        self.sudo_password = sudo_password
        self.slot_id = slot_id
        self.prompt = munch.Munch()

        # Prelogin prompts
        self.prompt.password_prompt = r'.*[Pp]assword: '
        self.prompt.prelogin_prompt = r'((%N )?(?<!Last )login: )'

        # MIO level prompts
        self.prompt.mio_prompt = r'%N([ /\w\-\*]+)?# '
        self.prompt.local_mgmt_prompt = r'%N\(local\-mgmt\)# '
        self.prompt.fxos_prompt = r'%N\(fxos\)# '
        self.prompt.fpr_module_prompt = r'\r\nFirepower-module.*>'
        self.prompt.cimc_prompt = r'\r\n\[.*\]# '

        self.prompt.rommon_prompt = r'rommon.*>' #?
        self.prompt.firepower_boot = r'-boot>' #?
        self.prompt.switch_boot = r'switch\(boot\)#' #?

        # FTD level prompts
        self.prompt.fireos_prompt = r'\r\n(\x1bE\x1b\[J)?> $'
        self.prompt.expert_cli = r'\r\n(\x1b\[18t)?admin@.*\$ $'
        self.prompt.sudo_prompt = '\r\n[^\r\n]*(root@.*#) $'
        self.prompt.disable_prompt = '[\r\n]{}> $'.format(app_hostname)
        self.prompt.enable_prompt = '[\r\n]{}# $'.format(app_hostname)
        self.prompt.config_prompt = '[\r\n]{}[\w]*\([\w\-]+\)# $'.format(app_hostname)

        # ASA level prompts
        self.prompt.asa_prompt = r'\r\nasa.*[>#]'

        logger.info('\n\n\nPrompts are: {}\n\n\n'.format(self.prompt))

    @property
    def hostname(self):
        return self.dev_hostname

    @hostname.setter
    def hostname(self, value):
        self.dev_hostname = value




import setuptools

with open("README.md", "r") as fh:
      long_description = fh.read()

setuptools.setup(
      name='firepower-kickstart',
      version='0.2.0',
      author="Cisco Systems Inc.",
      author_email="firepower-kickstart-support-ext@cisco.com",
      description="Python modules to install Firepower Threat Defense images (FTD) on hardware platforms.",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/CiscoDevNet/firepower-kickstart",
      packages=setuptools.find_packages(exclude=["*.tests", "*.unittests", "*.unittest", "*.sample_tests"]),
      install_requires=['pyVmomi', 'paramiko', 'unicon', 'boto3', 'munch', 'beautifulsoup4', 'Fabric3'],
      include_package_data=True,
      classifiers=[
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: Implementation :: CPython",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: POSIX :: Linux",
      ],
)



import munch


class ElektraPatterns:
    def __init__(self, hostname, login_password, sudo_password, enable_password):
        self.hostname = hostname
        self.username = 'admin'
        self.login_password = login_password
        self.sudo_password = sudo_password
        self.enable_password = enable_password
        self.default_password = 'Admin123'

        self.prompt = munch.Munch()

        self.prompt.password_prompt = '[\r\n][Pp]assword: '
        self.prompt.prelogin_prompt = '[\r\n]({} |firepower |asasfr |ciscoasa )login: '.format(self.hostname)
        self.prompt.fireos_prompt = r'[\r\n][\x07]?> $'
        self.prompt.expert_prompt = '[\r\n](admin@.*):~\$ $'
        self.prompt.sudo_prompt = '[\r\n](root@.*#) $'
        self.prompt.disable_prompt = '[\r\n]({}|firepower|ciscoasa)> $'.format(self.hostname)
        self.prompt.enable_prompt = '[\r\n]({}|firepower|ciscoasa)[/\w]*# $'.format(self.hostname)
        self.prompt.config_prompt = '[\r\n]({}|firepower|ciscoasa)[/\w]*\([\w\-]+\)# $'.format(self.hostname)
        self.prompt.rommon_prompt = 'rommon.*>'
        self.prompt.firepower_boot = '-boot>'



import logging

from kick.device2.fxos.utils import create_dict, scope

logger = logging.getLogger(__name__)


class ResourceProfile:
    def __init__(self, handle, name):
        """
        Constructor of Resource Profile
        :param handle: SSH connection handle
        :param name: resource profile name
        """
        self.handle = handle
        self.name = name

    def configure(self, cpu):
        """
        Configures resource profile with specified parameters
        :param cpu: cpu count
        """
        cpu = str(cpu)
        logger.info('Configuring Resource Profile {}'.format(self.name))
        scope(self.handle, 'ssa')
        self.handle.execute('enter resource-profile {}'.format(self.name))
        self.handle.execute('set cpu-core-count {}'.format(cpu))
        output = self.handle.execute('commit-buffer')
        if 'Error' in output:
            logger.info("Failed while trying to configure resource profile with error: {}".format(output))
            raise Exception("Failed while trying to configure resource profile")
        self.handle.execute('discard')

        output = create_dict(self.handle.execute('show detail'))

        assert output['Profile Name'] == self.name, \
            "Mismatch in Profile Name! Expected: {}, Saw: {}".format(self.name, output['Profile Name'])
        assert ('Physical CPU Count' in output and output['Physical CPU Count'] == cpu) or (
                'CPU Logical Core Count' in output and output['CPU Logical Core Count'] == cpu), \
            "Mismatch in CPU Count! Expected: {}, Saw: {}".format(cpu, output['Physical CPU Count'])
        logger.info('Successfully Configured Resource Profile {}'.format(self.name))

    def delete(self):
        """
        Deletes resource profile
        """
        logger.info('Deleting Resource Profile {}'.format(self.name))
        scope(self.handle, 'ssa')
        self.handle.execute('delete resource-profile {}'.format(self.name))
        output = self.handle.execute('commit')
        if 'Error' in output:
            logger.info("Failed while trying to delete resource profile with error: {}".format(output))
            raise Exception("Failed while trying to delete resource profile")
        self.handle.execute('discard')

        output = self.handle.execute('show resource-profile detail')
        assert 'Profile Name: {}\r'.format(self.name) not in output, \
            "Found Resource Profile When It Should've Been Deleted"
        logger.info('Successfully Deleted Resource Profile {}'.format(self.name))

    def get(self):
        """
        Get information about the resource profile
        :return: dictionary with the data
        """
        scope(self.handle, 'ssa')
        scope(self.handle, 'resource-profile {}'.format(self.name))
        resources = create_dict(self.handle.execute('show detail'))
        resources['name'] = resources.pop('Profile Name')
        resources['model'] = resources.pop('Security Model')
        resources['cpu_core_count'] = resources.pop('CPU Logical Core Count')
        resources['profile_type'] = resources.pop('Profile Type')
        resources['usage'] = resources.pop('Is In Use')

        return resources



import logging

from kick.device2.fxos.utils import create_dict, scope

logger = logging.getLogger(__name__)


class LogicalDeviceUtils:
    def __init__(self, handle):
        """
        Constructor of App Instance
        :param handle: SSH connection handle
        """
        self.handle = handle

    def get_all_ld(self):
        """
        Get all logical device details
        :return:
        """
        scope(self.handle, 'ssa')
        lds = []
        output = self.handle.execute('show logical-device detail', timeout=30)
        if "Logical Device" not in output:
            logger.info("Retrying command")
            output = self.handle.execute('show logical-device detail', timeout=30)
        if "Oper State" not in output:
            logger.info("Retrying command again")
            output = self.handle.execute('show logical-device detail', timeout=30)
        for i in list(
                map(lambda x: create_dict(x), output.split('\r\n\r\n'))):
            if hasattr(i, 'Name'):
                lds.append(i)

        return lds

    def delete_all_ld(self):
        """
        Delete all logical device
        :return:
        """
        all_devices = self.get_all_ld()
        for logical_device in all_devices:
            self.delete_ld_by_name(logical_device['Name'])

    def delete_ld_by_name(self, name):
        """
        delete logical device
        """
        logger.info('Deleting Logical-Device {}'.format(name))
        scope(self.handle, 'ssa')
        self.handle.execute('delete logical-device {}'.format(name))
        output = self.handle.execute('commit-buffer')
        if 'Error' in output:
            logger.info("Failed while deleting logical device with error: {}".format(output))
            raise Exception("Failed while deleting logical device")
        self.handle.execute('discard-buffer')



import re
import logging

from kick.device2.fxos.logical_device_handler import LogicalDeviceUtils
from kick.device2.fxos.utils import scope, create_dict
from unicon.eal.dialogs import Dialog
logger = logging.getLogger(__name__)


class LogicalDevice:
    def __init__(self, handle, name, app, slot, type):
        """
        :param handle: SSH connection handle
        :param name: logical device name; e.g. 'ftd1'
        :param app: the app associated to the logical device
        :param slot: the slot id
        :param type: logical device type; e.g. 'standalone' or clustered
        """
        self.handle = handle
        self.name = name
        self.app = app
        self.slot = slot
        self.type = type

    def create(self, interfaces, mgmt_bootstrap, cluster_bootstrap=None):
        """
        Create Logical Device
        """
        logger.info('Configuring Logical Device {}'.format(self.name))

        logger.info('Configure interfaces for app')
        for interface in interfaces:
            self.create_external_port_link(interface, commit=False)

        self.scope_to_ld()
        if cluster_bootstrap is not None:
            cluster_bootstrap.configure()
        mgmt_bootstrap.configure(slot=self.slot, app_name=self.app.app_name)

        self.scope_to_ld()
        logger.info('LD Configuration before Commit')
        self.handle.execute('show configuration')
        output = self.handle.execute('commit-buffer')
        if 'Error' in output:
            logger.info("Failed while creating logical device with error: {}".format(output))
            raise Exception("Failed while creating logical device")
        self.verify_ld()
        self.handle.execute('discard-buffer')

    def scope_to_ld(self):
        """
        Scope into ld configuration
        """
        scope(self.handle, 'ssa')
        self.handle.execute(
            'enter logical-device {} {} {} {}'.format(self.name, self.app.app_name, self.slot, self.type))

    def create_external_port_link(self, interface, commit=True, move_to_scope=True):
        """
        create external port link for interface
        :param interface: interface
        :param commit: if True it will commit buffer
        :param move_to_scope: True if we need to move to scope
        """
        if move_to_scope:
            self.scope_to_ld()
        intf_name = self.generate_int_name(interface.hardware)
        logger.info(
            'Configuring External Port Link Interface:{}\tname:{}\tapp:{}\n'.format(interface.hardware, intf_name,
                                                                                    self.app.app_name))
        self.handle.execute(
            'enter external-port-link {} {} {}'.format(intf_name, interface.hardware, self.app.app_name))
        self.handle.execute('up')
        if commit:
            output = self.handle.execute('commit-buffer')
            if 'Error' in output:
                logger.info("Failed while adding interface with error: {}".format(output))
                raise Exception("Failed while adding interface")

    def generate_int_name(self, hardware):
        """
        Generate interface name for external port link configuration
        """
        port_name = 'pc-' if 'Port-channel' in hardware else 'eth'
        return re.sub('/', '-', re.sub('Ethernet|Port-channel', port_name, hardware))

    def remove_external_port_link(self, interface, commit=True):
        """
        create external port link for interface
        :param interface: interface
        :param commit: if True it will commit buffer
        """
        self.scope_to_ld()
        intf_name = self.get_interface_name(interface.hardware)
        logger.info('Removing External Port Link Interface:{}\tname:{}\n'.format(interface.hardware, intf_name))
        output = self.handle.execute('delete external-port-link {}'.format(intf_name))
        if 'Error' in output:
            logger.info("Failed while removing external-port-link with error: {}".format(output))
            raise Exception("Failed while removing interface")
        if commit:
            output = self.handle.execute('commit-buffer')
            if 'Error' in output:
                logger.info("Failed while removing interface with error: {}".format(output))
                raise Exception("Failed while removing interface")

    def get_interface_name(self, hardware):
        """
        Get the interface name for external port link configuration
        """
        output = self.handle.execute('show external-port-link')
        pattern = '(\S+)\s+{}\s+.*'.format(hardware)
        match = re.search(pattern, output)
        if match:
            return match.group(1)
        else:
            raise Exception("Cannot determine the external-port name")

    def verify_ld(self):
        """
        Verify Logical Device is online
        :param handle: SSP handle
        :param ld_name: logical device name
        :return:
        """
        logger.info('Verifying LD Status')
        self.scope_to_ld()
        retry = 5
        while retry > 0:
            retry = retry - 1
            try:
                output = self.handle.execute('show detail', timeout=30)
                assert 'Oper State: Ok' in output, 'LD Oper state is not Ok'
                assert 'Switch Configuration Status: Ok' in output, 'LD Switch Configuration Status is not Ok'
                return
            except:
                logger.info("Retrying the command")
                pass
        raise Exception("The Logical device is not created")

    def delete(self):
        """
        delete logical device
        """
        logical_device_handler = LogicalDeviceUtils(self.handle)
        logical_device_handler.delete_ld_by_name(self.name)

    def cleanup_ld(self):
        """
        delete logical device and application
        """
        self.delete()
        self.app.delete()

    def get_ld_detail(self, move_to_scope=True):
        """
        Get logical device detail
        """
        if move_to_scope:
            self.scope_to_ld()
        return create_dict(self.handle.execute('show detail', timeout=30))



from unicon.statemachine import State, Path, StateMachine
from .patterns import WebserverPatterns


class WebserverStatemachine(StateMachine):
    def create(self):
        # Create States
        user_state = State('user_state', WebserverPatterns.user_prompt)
        # Add states
        self.add_state(user_state)

        # Create paths
        # Add paths




import logging
import re

log = logging.getLogger('AppArgumentsParser')
log.setLevel(logging.DEBUG)


class AppArgumentsParser:

    def get_apps_list(self, input: str):
        return_list = []
        apps_list = input.split(",")
        for app_range in apps_list:
            if "-" in app_range:
                if re.match("(\d+)\-(\d+)", app_range):
                    match = re.match("(\d+)\-(\d+)", app_range)
                    for i in range(int(match.groups()[0]), int(match.groups()[1]) + 1):
                        return_list.append("ftdapp{}".format(i))
                else:
                    log.error("The range {} is not valid. We support only digits.".format(app_range))
                    raise Exception("The range you specified is not valid")
            else:
                if re.match("^\d+$", app_range):
                    return_list.append("ftdapp{}".format(app_range))
                else:
                    return_list.append(app_range)

        return return_list

    def parse_cpu_core_count_argument(self, cpu_core_count: str):
        log.info("Parse cpu_core_count argument")
        resource_profiles = {}
        count_per_slot = {}
        i = 1
        for slot_cpu_info in cpu_core_count.split("|"):
            app_cpu_info = slot_cpu_info.split(",")
            for app_cpu in app_cpu_info:
                if app_cpu != "0" and app_cpu != "":
                    resource_profiles[app_cpu] = {}
                    resource_profiles[app_cpu]['name'] = 'stauto_{}'.format(app_cpu)
                    resource_profiles[app_cpu]['cpu_core_count'] = app_cpu
                else:
                    log.info("Cannot use {} for resource profile! Skipping...".format(app_cpu))
            count_per_slot[i] = app_cpu_info
            i = i + 1
        return {'count_per_slot': count_per_slot, 'resource_profiles': resource_profiles}

    def build_app_template(self, chassis_software, app_list):
        instances = chassis_software['applications']
        for i in instances:
            if (len(app_list) > 0 and instances[i].get("application_identifier") in app_list) or len(app_list) == 0:
                # need to initialize logical_device
                if not instances[i].get('logical_device'):
                    chassis_software['applications'][i]['logical_device'] = {}
                # if device_mode is not specified under logical_device it will take the value under chassis_software
                if not chassis_software['applications'][i]['logical_device'].get("device_mode"):
                    chassis_software['applications'][i]['logical_device']['device_mode'] = chassis_software.get(
                        'device_mode')
        if chassis_software.get('application_generic') is None:
            return chassis_software
        instance_template = chassis_software['application_generic']
        for i in instances:
            if (len(app_list) > 0 and instances[i].get("application_identifier") in app_list) or len(app_list) == 0:
                slot_id = None
                if not instances[i].get("slot"):
                    chassis_software['applications'][i]['slot'] = instance_template.get('slot')
                    slot_id = chassis_software['applications'][i]['slot']

                id = None
                if re.match("\D+(\d+)$", instances[i].get("application_identifier")):
                    match = re.match("\D+(\d+)$", instances[i].get("application_identifier"))
                    id = match.groups()[0]

                if not instances[i].get("application_name"):
                    chassis_software['applications'][i]['application_name'] = instance_template.get('application_name')
                if not instances[i].get("deploy_type"):
                    chassis_software['applications'][i]['deploy_type'] = instance_template.get('deploy_type')
                if not instances[i].get("resource_profile"):
                    chassis_software['applications'][i]['resource_profile'] = instance_template.get('resource_profile')

                # LOGICAL DEVICE
                if not instances[i].get('logical_device'):
                    chassis_software['applications'][i]['logical_device'] = {}
                if not instances[i]['logical_device'].get("name"):
                    chassis_software['applications'][i]['logical_device']['name'] = self._convert_name_with_id(
                        instance_template['logical_device'].get('name'), id, slot_id)

                # EXTERNAL PORT LINKS
                if not instances[i]['logical_device'].get('external_port_links'):
                    interface_list = []
                    for interface in instance_template['logical_device'].get("external_port_links"):
                        interface_list.append(self._convert_name_with_id(interface, id, slot_id))
                    chassis_software['applications'][i]['logical_device']['external_port_links'] = interface_list

                # BOOTSTRAP KEYS
                if not instances[i]['logical_device'].get('bootstrap_keys'):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys'] = {}
                if not instances[i]['logical_device']['bootstrap_keys'].get("firewall_mode"):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys']['firewall_mode'] = \
                        instance_template['logical_device']['bootstrap_keys'].get('firewall_mode')
                if not instances[i]['logical_device']['bootstrap_keys'].get("fqdn"):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys']['fqdn'] = \
                        self._convert_name_with_id(instance_template['logical_device']['bootstrap_keys'].get('fqdn'),
                                                   id, slot_id)
                if not instances[i]['logical_device']['bootstrap_keys'].get("dns_servers"):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys']['dns_servers'] = \
                        instance_template['logical_device']['bootstrap_keys'].get('dns_servers')
                if not instances[i]['logical_device']['bootstrap_keys'].get("search_domains"):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys']['search_domains'] = \
                        instance_template['logical_device']['bootstrap_keys'].get('search_domains')
                if not instances[i]['logical_device']['bootstrap_keys'].get("permit_expert_mode"):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys']['permit_expert_mode'] = \
                        instance_template['logical_device']['bootstrap_keys'].get('permit_expert_mode')
                if not instances[i]['logical_device']['bootstrap_keys'].get("firepower_manager_ip"):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys']['firepower_manager_ip'] = \
                        instance_template['logical_device']['bootstrap_keys'].get('firepower_manager_ip')
                if not instances[i]['logical_device']['bootstrap_keys'].get("management_type"):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys']['management_type'] = \
                        instance_template['logical_device']['bootstrap_keys'].get('management_type')

                # BOOTSTRAP KEYS SECRET
                if not instances[i]['logical_device'].get('bootstrap_keys_secret'):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys_secret'] = {}
                if not instances[i]['logical_device']['bootstrap_keys_secret'].get("password"):
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys_secret']['password'] = \
                        instance_template['logical_device']['bootstrap_keys_secret'].get('password')

                # IPv4
                if not instances[i]['logical_device'].get('ipv4'):
                    chassis_software['applications'][i]['logical_device']['ipv4'] = {}
                if not instances[i]['logical_device']['ipv4'].get("ip"):
                    chassis_software['applications'][i]['logical_device']['ipv4']['ip'] = \
                        self._compute_ipv4_address(instance_template['logical_device']['ipv4'].get('ip'), id)
                if not instances[i]['logical_device']['ipv4'].get("netmask"):
                    chassis_software['applications'][i]['logical_device']['ipv4']['netmask'] = \
                        instance_template['logical_device']['ipv4'].get('netmask')
                if not instances[i]['logical_device']['ipv4'].get("gateway"):
                    chassis_software['applications'][i]['logical_device']['ipv4']['gateway'] = \
                        instance_template['logical_device']['ipv4'].get('gateway')

                # IPv6
                if not instances[i]['logical_device'].get('ipv6'):
                    chassis_software['applications'][i]['logical_device']['ipv6'] = {}
                if not instances[i]['logical_device']['ipv6'].get("ip"):
                    chassis_software['applications'][i]['logical_device']['ipv6']['ip'] = \
                        self._compute_ipv6_address(instance_template['logical_device']['ipv6'].get('ip'), id)
                if not instances[i]['logical_device']['ipv6'].get("prefix"):
                    chassis_software['applications'][i]['logical_device']['ipv6']['prefix'] = \
                        instance_template['logical_device']['ipv6'].get('prefix')
                if not instances[i]['logical_device']['ipv6'].get("gateway"):
                    chassis_software['applications'][i]['logical_device']['ipv6']['gateway'] = \
                        instance_template['logical_device']['ipv6'].get('gateway')

                # Update data if deploy_type == native
                if chassis_software['applications'][i]['deploy_type'] == 'native':
                    chassis_software['applications'][i]['logical_device']['bootstrap_keys']['permit_expert_mode'] = None
                    chassis_software['applications'][i]['resource_profile'] = None

        return chassis_software

    def _convert_name_with_id(self, name, id, slot_id):
        if "{id}" in name:
            name = name.replace("{id}", id)
        if "{slot_id}" in name:
            name = name.replace("{slot_id}", str(slot_id))
        return name

    def _compute_ipv4_address(self, start_address, id):
        match = re.match("(\d+\.\d+\.\d+\.)(\d+)", start_address)
        address = int(match.groups()[1]) + int(id) - 1
        return "{}{}".format(match.groups()[0], address)

    def _compute_ipv6_address(self, start_address, id):
        match = re.match("([\w+\:]+\:)(\d+)$", start_address)
        address = int(match.groups()[1]) + int(id) - 1
        return "{}{}".format(match.groups()[0], address)



from unicon.eal.dialogs import Statement


def password_handler(spawn, password):
    spawn.sendline(password)


def enable_handler(spawn, password):
    spawn.sendline('enable')
    spawn.expect('Password: ')
    spawn.sendline(password)


class Ftd5500xStatements:
    def __init__(self, patterns):
        self.login_password_statement = Statement(pattern=patterns.prompt.password_prompt,
                                                  action=password_handler,
                                                  args={'password': patterns.login_password},
                                                  loop_continue=True,
                                                  continue_timer=True)

        self.sudo_password_statement = Statement(pattern=patterns.prompt.password_prompt,
                                        action=password_handler,
                                        args={'password': patterns.sudo_password},
                                        loop_continue=True,
                                        continue_timer=True)

        self.enable_password_statement = Statement(pattern=patterns.prompt.disable_prompt,
                                              action=enable_handler,
                                              args={'password': patterns.enable_password},
                                              loop_continue=True,
                                              continue_timer=True)



import logging

from kick.device2.fxos.utils import create_dict, scope

logger = logging.getLogger(__name__)


class AppInstanceUtils:
    def __init__(self, handle, slot):
        """
        Constructor of App Instance
        :param handle: SSH connection handle
        :param slot: slot id; e.g. '1', '2'
        """
        self.handle = handle
        self.slot = str(slot)
        if self.slot == "":
            raise ValueError('Slot is mandatory and was sent as empty string')

    def show_app_instance(self):
        """
        Show Application Instance: show app-instance
        :return:
        """
        self.handle.execute("top")
        scope(self.handle, 'ssa/slot {}'.format(self.slot))
        out = self.handle.execute('show app-instance')
        if 'Error' in out:
            logger.info("Failed while trying to show app-instance with error: {}".format(out))
            raise Exception("Failed while trying to show app-instance")
        return out

    def show_app_instance_detail(self):
        """
        Show Application Instance details: show app-instance detail
        :param slot: slot id
        :return:
        """
        self.handle.execute("top")
        scope(self.handle, 'ssa/slot {}'.format(self.slot))
        out = self.handle.execute('show app-instance detail')
        if 'Error' in out and 'Error Msg' not in out:
            logger.info("Failed while trying to show app-instance detail with error: {}".format(out))
            raise Exception("Failed while trying to show app-instance detail")
        return out

    def get_app_instances_in_slot(self):
        """
        Get app-instance in slot as munch dictionary
        :return: Dictionary structure with apps
        """
        apps = []
        scope(self.handle, 'ssa/slot {}'.format(self.slot))
        for i in list(map(lambda x: create_dict(x), self.handle.execute('show app-instance detail').split('\r\n\r\n'))):
            if hasattr(i, 'App Name'):
                apps.append(i)
        return apps

    def get_online_app_instances_in_slot(self):
        """
        Get the list of online app-instance in slot
        :return: List of app identifiers
        """
        apps = []
        scope(self.handle, 'ssa/slot {}'.format(self.slot))
        for i in list(map(lambda x: create_dict(x), self.handle.execute('show app-instance detail').split('\r\n\r\n'))):
            if hasattr(i, 'App Name'):
                if hasattr(i, 'Oper State') and i['Oper State'] == 'Online':
                    apps.append(i['Identifier'])
        return apps

    def show_monitor_detail(self):
        self.handle.execute("top")
        scope(self.handle, 'ssa/slot {}'.format(self.slot))
        out = self.handle.execute('show monitor detail')
        if 'Error' in out:
            logger.info("Failed while trying to show monitor detail with error: {}".format(out))
            raise Exception("Failed while trying to show app-instance")
        resources = create_dict(out)
        resources['available_cores'] = resources.pop('CPU Cores Available')

        return resources



import configparser
import os

config = configparser.ConfigParser()
try:
    kick_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    config.read(os.path.join(kick_path, 'config.ini'))
except:
    pass

DEFAULT = 'DEFAULT'
USERNAME = 'username'
PASSWORD = 'password'
EN_PASSWORD = 'en_password'

def get_username(user):
    """
    Gets the username from config file if present

    :param user: username given by user
            if not given, defaulted to ['myusername']
    :return: username
    """
    if USERNAME in config[DEFAULT]:
        return config[DEFAULT][USERNAME]
    else:
        return user


def get_password(pwd):
    """
    Gets the password from config file if present

    :param pwd: password given by user
            if not given, defaulted to ['mypassword']
    :return: password
    """
    if PASSWORD in config[DEFAULT]:
        return config[DEFAULT][PASSWORD]
    else:
        return pwd


def get_en_password(en_password):
    """
    Gets the en_password from config file if present

    :param en_password: en_password given by user
            if not given, defaulted to ['myenpassword']
    :return: en_password
    """
    if EN_PASSWORD in config[DEFAULT]:
        return config[DEFAULT][EN_PASSWORD]
    else:
        return en_password


class KickConsts(object):
    DEFAULT_PASSWORD = 'Admin123'
    DUMMY_PASSWORD = 'Admin123!'



import munch


class Series3Patterns:
    """Series3 FTD class for device prompt patterns."""
    def __init__(self, hostname, login_username, login_password, sudo_password):
        """Constructor for Series3 Patterns."""

        self.hostname = hostname
        self.login_username = login_username
        self.login_password = login_password
        self.sudo_password = sudo_password
        self.username = 'admin'
        self.default_password = 'Admin123'
        self.prompt = munch.Munch()

        self.prompt.password_prompt = r'.*Password: '
        #self.prompt.prelogin_prompt = r'({} )login: '.format(self.hostname)
        self.prompt.prelogin_prompt = '\r?\n?({}|firepower) login: '.format(self.hostname)
        self.prompt.fireos_prompt = '\n[\x07]?>'
        self.prompt.expert_prompt = '\n(admin@(firepower|{})*):~\$ '.format(self.hostname)
        self.prompt.sudo_prompt = r'\nroot@(firepower|{}).*#'.format(self.hostname)
        self.prompt.eula_prompt = 'You must accept the EULA to continue'
        self.prompt.firstboot_prompt = 'You must change the password for '
        self.prompt.lilo_boot_prompt = 'boot:'
        self.prompt.lilo_boot_menu_prompt = 'qqqqqqqqqqqqqqqqqqqqqqqqqqqqqq'
        self.prompt.lilo_os_prompt = "root@\(none\):.*#"

